@@ -6,6 +6,8 @@ import re
 from collections import OrderedDict, defaultdict, deque, namedtuple
 
 import ufoLib2
+from fontTools.misc.arrayTools import Vector
+from fontTools.misc.fixedTools import otRound
 from fontTools.ttLib import TTFont, TTLibError, newTable
 from fontTools.ttLib.tables._g_l_y_f import (
     ROUND_XY_TO_GRID,
@@ -14,6 +16,8 @@ from fontTools.ttLib.tables._g_l_y_f import (
     USE_MY_METRICS,
 )
 from fontTools.ttLib.tables.ttProgram import Program
+from fontTools.ttLib.tables.TupleVariation import TupleVariation
+from fontTools.varLib import models
 
 import vttLib.transfer
 from vttLib.parser import AssemblyParser, ParseException
@@ -706,14 +710,58 @@ def update_composites(font, glyphs=None, vtt_version=6):
         set_glyph_assembly(font, glyph_name, new_data)
 
 
-def compile_instructions(font, ship=True):
+def compile_instructions(font, ship=True, keep_cvar=False):
     if "glyf" not in font:
         raise VTTLibError("Missing 'glyf' table; not a TrueType font")
     if "TSI1" not in font:
         raise VTTLibError("The font contains no 'TSI1' table")
+    if keep_cvar and "cvar" not in font:
+        raise VTTLibError(
+            "The keep_cvar parameter is set, but the cvar table is missing from the font"
+        )
 
     control_program = get_extra_assembly(font, "cvt")
     set_cvt_table(font, control_program)
+    if "TSIC" in font and not keep_cvar:
+        # Compile TSIC
+        tsic = font["TSIC"].table
+        cvts = font["cvt "].values
+        # Gather the full set of cvts (not just those in TSIC) for all locations
+        cvt_sets = [Vector(cvts)]
+        for record in tsic.Record:
+            cvt_local = {i: v for i, v in zip(record.CVTArray, record.CVTValueArray)}
+            cvt_set = [
+                cvt_local[i] if i in cvt_local else cvt for i, cvt in enumerate(cvts)
+            ]
+            cvt_sets.append(Vector(cvt_set))
+
+        # Compute variations
+        locs = [{axis_tag: 0.0 for axis_tag in tsic.AxisArray}]  # default instance
+        locs += [
+            {axis_tag: val for axis_tag, val in zip(tsic.AxisArray, loc.Axis)}
+            for loc in tsic.RecordLocations
+        ]
+        model = models.VariationModel(locs)
+        variations = []
+        deltas, supports = model.getDeltasAndSupports(cvt_sets)
+        for i, (delta, support) in enumerate(zip(deltas[1:], supports[1:])):
+            delta = [otRound(d) for d in delta]
+            if all(abs(v) <= 0.5 for v in delta):
+                continue
+            # Remove any deltas that weren't specified in TSIC at this location
+            # TODO: Just replace optimizaton with getting rid of 0 deltas?
+            tsic_rec_index = model.reverseMapping[i + 1] - 1  # Map back to TSIC records
+            delta = [
+                d if j in tsic.Record[tsic_rec_index].CVTArray and d != 0 else None
+                for j, d in enumerate(delta)
+            ]
+            var = TupleVariation(support, delta)
+            variations.append(var)
+
+        if variations:
+            cvar = font["cvar"] = newTable("cvar")
+            cvar.version = 1
+            cvar.variations = variations
 
     for tag in ("prep", "fpgm"):
         if tag not in font:
@@ -814,7 +862,7 @@ def vtt_dump_file(infile, outfile=None, **_):
     vttLib.transfer.dump_to_file(font, outfile)
 
 
-def vtt_merge_file(infile, outfile=None, **_):
+def vtt_merge_file(infile, outfile=None, keep_cvar=False, **_):
     """Write VTT data from a TTX dump to a TTF."""
     if not os.path.exists(infile):
         raise vttLib.VTTLibArgumentError("Input file '%s' not found" % infile)
@@ -823,7 +871,7 @@ def vtt_merge_file(infile, outfile=None, **_):
         raise vttLib.VTTLibArgumentError("Output file '%s' not found" % outfile)
 
     font = TTFont(outfile)
-    vttLib.transfer.merge_from_file(font, infile)
+    vttLib.transfer.merge_from_file(font, infile, keep_cvar=keep_cvar)
     font.save(outfile)
 
 
@@ -843,13 +891,24 @@ def vtt_move_ufo_data_to_file(infile, outfile=None, **_):
 
 
 def vtt_compile(
-    infile, outfile=None, ship=False, inplace=None, force_overwrite=False, **_
+    infile,
+    outfile=None,
+    ship=False,
+    inplace=None,
+    force_overwrite=False,
+    keep_cvar=False,
+    **_,
 ):
     if not os.path.exists(infile):
         raise vttLib.VTTLibArgumentError("Input TTF '%s' not found." % infile)
 
     font = TTFont(infile)
 
+    if keep_cvar and "cvar" not in font:
+        raise vttLib.VTTLibArgumentError(
+            "The --keep-cvar option was specified, but no cvar table exists in the font"
+        )
+
     if outfile:
         pass
     elif inplace:
@@ -867,5 +926,5 @@ def vtt_compile(
 
         outfile = makeOutputFileName(infile, None, ".ttf")
 
-    vttLib.compile_instructions(font, ship=ship)
+    vttLib.compile_instructions(font, ship=ship, keep_cvar=keep_cvar)
     font.save(outfile)
